/*
===============================================================================
 Name        : MP3_16.c
 Author      : $(author)
 Version     :
 Copyright   : $(copyright)
 Description : main definition
===============================================================================
*/

#ifdef __USE_CMSIS
#include "LPC17xx.h"
#endif

#include <cr_section_macros.h>
#include "stdlib.h"
#include "string.h"
//#include "ssp.h"
#include "i2s.h"
#include "timer.h"
//#include "i2c.h"
//#include "type.h"
// libmad
#include "libmad/config.h"
#include "libmad/decoder.h"
// FatFS
#include "usbhost_inc.h"
#include "ff.h"
#include "ffconf.h"
#include "uart.h"
// 自己／デバッグ
#include "mp3_16.h"

#define SAMPLE_FREQ 			44100	// サンプリング周波数の初期値

#define INPUT_BUF_SIZE			(512*4)		// madlibのinput関数内バッファのサイズ 	2100

#define TIMER0_INTERVAL			((SystemCoreClock/(4*1000*10)))			// 0.1ms (PCLK=CCLK/4)
#define TIMER1_INTERVAL			((SystemCoreClock/(4*5)))				// 200ms=1/5s (PCLK=CCLK/4)
#define TIMER1_1MIN				(5*60)		// 1分
#define TIMER2_INTERVAL			(SystemCoreClock/4)						// 1s (PCLK=CCLK/4)
#define	TIMER2_PRESCALE			20

// ボタン操作
#define SW_PRESSED				0x01
#define SWPLAY_PRESSED          0x01
#define SWOPTION_PRESSED     	0x02
#define SWNEXT_PRESSED          0x03
#define SWDESELECT_PRESSED     	0x04
#define SWPREV_PRESSED          0x05
#define SWUP_PRESSED            0x06
#define SWSELECT_PRESSED      	0x07
#define SWDOWN_PRESSED          0x08
#define SWPOFF_PRESSED			0x09

#define POWER_ON_BIT			31		// P1.31
#define SPEAKER_MUTE_BIT		26		// P0.26
#define MCP23017_RESET_PORT		 6		// P0.6
#define POWER_LED_BIT			24		// P0.24

#define DEBUG_MSG   DebugWriteln

// FatFs
FATFS Fatfs[_VOLUMES];		/* File system object for each logical drive */
uint8_t Drives;				/* Drives mask */
//static FATFS        *pFatfs;	// MP3_12
//static FIL 			*pfilMP3;
FIL 			FilMP3;

// DAC/madlib
typedef unsigned short		dacout_s;
dacout_s dacbuf[1152*4];
dacout_s *dac_s, *dac_e;
dacout_s *dac_s2, *dac_e2;
dacout_s *buftab[2];
int 	dacflg;
void	*pDecoderSync;		// decoder->sync 退避 (decoder.c)
void	*pStreamMainData;	// stream->main_data退避 (layer3.c)
void	*pFrameOverlap;		// frame->overlap退避 (layer3.c)
unsigned int samplerate;	// サンプルレートの退避
enum mad_flow (*output_func_backup)(void *, struct mad_header const *, struct mad_pcm *);	// output_funcの一時的退避

// イベント/制御
static volatile int	PinEvent;
static uint8_t      PonRestart;     // execute Power on Restart
static volatile uint16_t	CHATTERING_TMR;
static volatile uint16_t	AutoPoffTimer;
// 9/23
//static uint8_t    	LcdImageBuff[LCD_IMAGE_BUFFER_SIZE]; 	// LCDフル画面イメージデータ
//static uint8_t      PlayDirHist[RECLEN_HISTFILE];			// 履歴ファイルレコード
static uint8_t		SpeakerOn;		// 1=ON, 0=OFF
static uint8_t		ChannelMode;	// 1=STEREO, 0=MONO
static uint8_t		NowPlaying;

typedef enum {PM_RANDOM_DIR, PM_RANDOM_TRACK, PM_SELECT} playmode_t;
static playmode_t   PlayMode;

typedef enum {PS_PLAY, PS_SELECT} playorselect_t;

// 音量
volatile static mad_fixed_t		VolumeFactor;
volatile static uint16_t		Volume;
static uint16_t		VolumeTab[] = {8,10,13,16,20,25,32,40,50,63,
									79,100,126,158,199,251,316,398,501,630,
									793,999,1258,1583,1994,2510,3160,3978,5009,6306};	// MAX_VOLUME=10000
#define	MIN_VOLUMETAB_IDX		0
#define MAX_VOLUMETAB_IDX		((sizeof(VolumeTab)/sizeof(VolumeTab[0])) - 1)
#define	DEF_VOLUMETAB_IDX		20		// 15 default
#define MIN_VOLUME              0		// Volume最小値
#define MAX_VOLUME              10000	// Volume最大値

// 関数宣言
static int decode();
int PauseAndSelectItem();
void SysTick_Off();
void SysTick_On();
void SystemInit (void);
void SystemClockUpdate (void);
void SystemCoreClockUpdate (void);
uint8_t Mem_Init (void);
int USBHost_Init(void);
void InitApp();
int PlayOneMp3(const char *mp3file, uint16_t dirno, uint8_t trackno);
//void test_read_write();


int main(void) {
    // Force the counter to be placed into memory
    volatile static int i = 0 ;

    SystemInit();
    SystemCoreClockUpdate();
    InitApp();
    CHATTERING_TMR = 0;
    AutoPoffTimer = 0;
    NowPlaying = 0;

//    test_read_write();

    PlayOneMp3("/MSC1_in/b.mp3", 0, 0);
    PlayOneMp3("/a.mp3", 0, 0);

    // Enter an infinite loop, just incrementing a counter
    while(1) {
        i++ ;
    }
    return 0 ;
}

/*
 * 各種初期化(IO/Timer/I2S/SPI/I2C/UART/GCLD/メモリ領域)
 */
void InitApp(){
    LPC_GPIO1->FIODIR |= ((0x01 << 18) | (0x01 << 20) | (0x01 << 21) | (0x01 << 23));		// on board led
    LPC_GPIO1->FIOSET |= (0x01 << 18);
    UARTInit(1, 9600);
	DEBUG_MSG("MP3_16 start.\n\r");

	//pFatfs = (FATFS *)mad_malloc(sizeof(FATFS));
	//pfilMP3 = (FIL *)mad_malloc(sizeof(FIL));

    Drives = Mem_Init();
    if (Drives == 0)
    {
        DEBUG_MSG ("Memory: Init failed.\n\r");
        //LCD_PrintString(0*8, 7*15, "Mem: Init Fail", COLOR_RED, COLOR_BLACK);
        //goto loop;
        return;
    }
    LPC_GPIO1->FIOSET |= (0x01 << 20);

	init_timer( 0, TIMER0_INTERVAL );
	init_timer( 1, TIMER1_INTERVAL );
	i2s_init();		// i2s.c
	enable_timer( 0 );
	enable_timer( 1 );

	pDecoderSync = 0;
	pStreamMainData = 0;
	pFrameOverlap = 0;

	dac_s = dac_e = dacbuf;
	dac_s2 = dac_e2 = dacbuf;
	buftab[0] = dacbuf;
	buftab[1] = &dacbuf[1152*2];
	dacflg = 0;

	//VolumeFactor = MAD_F_ONE;		//mad_f_tofixed((long)options.volume/100.0);
	Volume = VolumeTab[DEF_VOLUMETAB_IDX];
	VolumeFactor = mad_f_tofixed((long)Volume/(float)MAX_VOLUME);
	//DecodeSkipflg = DECODE_SKIP_OUT;

	samplerate = SAMPLE_FREQ;
	i2s_begin(samplerate, 16);

	PinEvent = 0;
}

/**
  * @brief  Initializes memory.
  *
  * @param  None.
  * @retval Mask of mounted drives
  *
  */
uint8_t Mem_Init (void)
{
    uint8_t retv = 0x0;

	/* Wait until there is USB flash disk inserted. */
    //LCD_PrintString(0*8, 4*15, "Connect USB MSD", COLOR_WHITE, COLOR_BLACK);
	if (USBHost_Init() == OK)
	{
		DEBUG_MSG("USBHost: Init OK.\n\r");
		//LCD_PrintString(0*8, 5*15, "USB: Init OK", COLOR_LIME, COLOR_BLACK);
		if (f_mount(/*pFatfs*/ &Fatfs[0], _T("0:"), 0) == FR_OK){
			DEBUG_MSG("USBFS: Mount OK.\n\r");
			//LCD_PrintString(0*8, 6*15, "USBFS: Mount OK", COLOR_LIME, COLOR_BLACK);
			retv |= 1;
		} else {
			DEBUG_MSG("USBFS: Mount failed.\n\r");
			//LCD_PrintString(0*8, 6*15, "USBFS: mnt fail", COLOR_RED, COLOR_BLACK);
	    }
	}
	else
	{
		DEBUG_MSG("USBHost: Init failed.\n\r");
		//LCD_PrintString(0*8, 5*15, "USB: Init fail", COLOR_RED, COLOR_BLACK);
		return retv;
	}

#ifdef SUPPORT_SD
	if (f_mount(1, &Fatfs[1]) == FR_OK) {
		DEBUG_MSG("SDFS: Mount OK.\n\r");
		LCD_PrintString(0*8, 6*15, "SDFS: Mount OK", COLOR_LIME, COLOR_BLACK);
		retv |= 2;
	} else {
		DEBUG_MSG("SDFS: Mount failed.\n\r");
		LCD_PrintString(0*8, 6*15, "SDFS: mnt fail", COLOR_RED, COLOR_BLACK);
    }
#endif

    return (retv);
}

/*----------------------------------------------------------------------------
  USB functions
 *----------------------------------------------------------------------------*/
/**
  * @brief  Initializes the USBHost .
  *
  * @param  None
  * @retval 0:OK, others: failure code
  *
  */
int USBHost_Init(void)
{
    int32_t  rc;
	uint32_t  numBlks, blkSize;
	uint8_t  inquiryResult[INQUIRY_LENGTH];

    Host_Init();               /* Initialize the lpc17xx host controller */
    rc = Host_EnumDev();       /* Enumerate the device connected  */
	if (rc != OK)
		return 1;
	rc = MS_Init( &blkSize, &numBlks, inquiryResult );
	if ((rc != OK) || (blkSize != 512))
		return 2;

	return OK;
}


/*
 * MP3再生
 * 	mp3file:MP3ファイルフルパス、dirno:ルートディレクトリ=0から始まる行番号
 */
int PlayOneMp3(const char *mp3file, uint16_t dirno, uint8_t trackno){
    FRESULT     res;
    int         ret;
//    char		buff[INPUT_BUF_SIZE];
//    UINT		bytes;

    ret = FR_OK;
	DEBUG_MSG("MP3_16 start.\n\r");
	DEBUG_MSG((char *)(mp3file));
	res = f_open(&FilMP3 /*pfilMP3*/, mp3file, FA_READ | FA_OPEN_EXISTING);
	if (res != FR_OK){
		DEBUG_MSG("f_open ERR.\n\r");
		return res;
	}
	DEBUG_MSG("f_open OK.\n\r");
//	do {
//		res = f_read(&FilMP3, buff, INPUT_BUF_SIZE, &bytes);
//		if (res != FR_OK){
//			DEBUG_MSG("f_read err.\n\r");
//			break;
//		} else if (bytes == 0){
//			break;
//		}
//		DEBUG_MSG("f_read OK.\n\r");
//	} while (1);
//	DEBUG_MSG("f_read ALL OK.\n\r");
	//DebugWriteln((char *)"f_open OK");

	// 9/23
//	if (PonRestart == 1){       // POWER ON RESTART
//		DWORD	lastbyte;
//		memcpy(&lastbyte, PlayDirHist + HS_LASTBYTE, 4);
//		PonRestart = 0;
//		if (f_lseek(pfilMP3, lastbyte) != FR_OK){      // go to restart point
//			f_lseek(pfilMP3, 0);
//		}
//	}
//	//f_lseek(pfilMP3, 4*32*100000L);		// 途中から再生テスト
//
//    memcpy(PlayDirHist + HS_LASTDIR, &dirno, 2);
//    *(uint16_t *)(PlayDirHist + HS_LASTTRACK) = trackno;
//    *(uint32_t *)(PlayDirHist + HS_LASTBYTE) = f_tell(pfilMP3);

	i2s_init();		// i2s.c
	dac_s = dac_e = dacbuf;
	dac_s2 = dac_e2 = dacbuf;
	buftab[0] = dacbuf;
	buftab[1] = &dacbuf[1152*2];
	dacflg = 0;

	NowPlaying = 1;
	ret = decode();
	NowPlaying = 0;

// 9/23
//    *(uint32_t *)(PlayDirHist + HS_LASTBYTE) = f_tell(pfilMP3);
	f_close(&FilMP3 /*pfilMP3*/);

// 9/23
//	//if (ret == GOTO_MAINMENU || ret == SELECT_OTHER_MP3){	// 中断のときだけ履歴保存
//		SaveHistory(pfilMP3);
//	//}

	return ret;
}

/*
 * This is a private message structure. A generic pointer to this structure
 * is passed to each of the callback functions. Put here any data you need
 * to access from within the callbacks.
 */
struct buffer {
  unsigned char const *start;
  unsigned long length;
};


/*
 * This is the input callback. The purpose of this callback is to (re)fill
 * the stream buffer which is to be decoded. In this example, an entire file
 * has been mapped into memory, so we just call mad_stream_buffer() with the
 * address and length of the mapping. When this callback is called a second
 * time, we are finished decoding.
 */
static
enum mad_flow input(void *data,
		    struct mad_stream *stream)
{
	static unsigned char strmbuff[INPUT_BUF_SIZE];
	int ret;
	int rsz;
	unsigned char *bp;
	UINT	bytes;
	//int		i;
	//uint16_t	vstep;
	//DEBUG_MSG("input.\n\r");

	bp = strmbuff;
	rsz = 0;
	if(stream->error == MAD_ERROR_BUFLEN || stream->buffer == 0 /*NULL*/)
	{
	if(stream->next_frame != 0 /*NULL*/)
	  {
		 rsz = stream->bufend - stream->next_frame;
//		 memmove(pStrmBuff, stream->next_frame, rsz);
		 memmove(strmbuff, stream->next_frame, rsz);
		 bp = strmbuff + rsz;
//		 bp = pStrmBuff + rsz;
	  }
	}
	//struct buffer *buffer = data;
	  //if (!buffer->length)

	//DebugWritelnV1("read bytes.\n\r", INPUT_BUF_SIZE - rsz);
	ret = f_read(&FilMP3 /*pfilMP3*/, bp, INPUT_BUF_SIZE - rsz, &bytes);
	if (ret != FR_OK || bytes == 0){
		DebugWritelnV1("f_read err.\n\r", ret);
		return MAD_FLOW_STOP;
	}
	//DEBUG_MSG("f_read.\n\r");

	mad_stream_buffer(stream, strmbuff, bytes + rsz);
//	mad_stream_buffer(stream, pStrmBuff, bytes + rsz);
	//mad_stream_buffer(stream, buffer->start, buffer->length);

	//buffer->length = 0;

	if (AutoPoffTimer >= 180){
		PinEvent = 0;
		return MAD_FLOW_POWEROFF;
	}
	if (PinEvent > 0){
		switch (PinEvent){
		case SWNEXT_PRESSED:
			PinEvent = 0;
			return MAD_FLOW_NEXT;
			break;
		case SWPREV_PRESSED:
			PinEvent = 0;
			return MAD_FLOW_PREV;
			break;
		case SWPLAY_PRESSED:
#if 0
			if (DecodeSkipflg == DECODE_SKIP_OUT){		//曲の途中の無音状態を探してそこまでスキップする。マイコンのスピードが不足して実用的でない。
				DecodeSkipflg = DECODE_SKIP_BEGIN;
				SkipCounter = 0;
			}
#endif
			PinEvent = 0;
			return MAD_FLOW_PLAY;
			break;
		case SWDESELECT_PRESSED:
			PinEvent = 0;
			return MAD_FLOW_DESELECT;
			break;
		case SWOPTION_PRESSED:
#if 0
			InitGlcdFull();
			DispLcdFullImage();
#endif
			/*
			if (glcd_disp_status == 0){
				GLCD_DISP_ON();
			} else {
				GLCD_DISP_OF();
			}
			*/
			break;
		case SWPOFF_PRESSED:
			PinEvent = 0;
			return MAD_FLOW_POWEROFF;
			break;
		}
		PinEvent = 0;
	}
	return MAD_FLOW_CONTINUE;
}

/*
 * The following utility routine performs simple rounding, clipping, and
 * scaling of MAD's high-resolution samples down to 16 bits. It does not
 * perform any dithering or noise shaping, which would be recommended to
 * obtain any exceptional audio quality. It is therefore not recommended to
 * use this routine if high-quality output is desired.
 * mpg321-0.3.2.origのaudio_linear_ditherを使ってみた。
 */
static inline
signed int scale(mad_fixed_t sample)
{
  /* round */
  sample += (1L << (MAD_F_FRACBITS - 16));

  /* clip */
  if (sample >= MAD_F_ONE)
    sample = MAD_F_ONE - 1;
  else if (sample < -MAD_F_ONE)
    sample = -MAD_F_ONE;

  /* quantize */
  return sample >> (MAD_F_FRACBITS + 1 - 16);
}

/*
 * mpg321-0.3.2.origから引用
 */
struct audio_dither {
  mad_fixed_t error[3];
  mad_fixed_t random;
};

/*
 * NAME:        prng()
 * DESCRIPTION: 32-bit pseudo-random number generator
 * mpg321-0.3.2.origから引用
 */
/* 9/23 static*/ inline
unsigned long prng(unsigned long state)
{
  return (state * 0x0019660dL + 0x3c6ef35fL) & 0xffffffffL;
}

/*
 * NAME:        audio_linear_dither()
 * DESCRIPTION: generic linear sample quantize and dither routine
 * mpg321-0.3.2.origから引用
 */
inline
signed long audio_linear_dither(unsigned int bits, mad_fixed_t sample,
                                struct audio_dither *dither)
{
  unsigned int scalebits;
  mad_fixed_t output, mask, random;

  enum {
    MIN = -MAD_F_ONE,
    MAX =  MAD_F_ONE - 1
  };

  /* noise shape */
  sample += dither->error[0] - dither->error[1] + dither->error[2];

  dither->error[2] = dither->error[1];
  dither->error[1] = dither->error[0] / 2;

  /* bias */
  output = sample + (1L << (MAD_F_FRACBITS + 1 - bits - 1));

  scalebits = MAD_F_FRACBITS + 1 - bits;
  mask = (1L << scalebits) - 1;

  /* dither */
  random  = prng(dither->random);
  output += (random & mask) - (dither->random & mask);

  dither->random = random;

  /* clip */
  if (output > MAX) {
    output = MAX;

    if (sample > MAX)
      sample = MAX;
  }
  else if (output < MIN) {
    output = MIN;

    if (sample < MIN)
      sample = MIN;
  }

  /* quantize */
  output &= ~mask;

  /* error feedback */
  dither->error[0] = sample - output;

  /* scale */
  return output >> scalebits;
}

/*
 * This is the output callback function. It is called after each frame of
 * MPEG audio data has been completely decoded. The purpose of this callback
 * is to output (or play) the decoded PCM audio.
 * ステレオ用
 */
static
enum mad_flow output(void *data,
		     struct mad_header const *header,
		     struct mad_pcm *pcm)
{
  unsigned int nsamples;
  mad_fixed_t const *left_ch, *right_ch;
  static struct audio_dither dither;
  dacout_s *dac_w;

  /* pcm->samplerate contains the sampling frequency */
  //nchannels = pcm->channels;
  nsamples  = pcm->length;
  left_ch   = pcm->samples[0];
  right_ch  = pcm->samples[1];

  // dac_s2が空くのを待つ(SysTick_Handler参照)
  while(dac_s2 < dac_e2){
	  Pause();
  }

  dac_w = buftab[dacflg];
  dac_s2 = dac_w;
  while (nsamples--) {
		*dac_w = (unsigned short)audio_linear_dither(16, mad_f_mul(*left_ch, VolumeFactor), &dither);
	    dac_w++;
    	*dac_w = (unsigned short)audio_linear_dither(16, mad_f_mul(*right_ch, VolumeFactor), &dither);;
	    dac_w++;
	    left_ch++;
	    right_ch++;
  }
  dac_e2 = dac_w;
  dacflg = 1 - dacflg;

#if 0		// madplayerの原型
//  LPC_GPIO1->FIOSET |= (0x01 << 18);
	enable_timer( 0 );
  while(dac_s < dac_e){
	  Pause();
	  //int k;
	  //for (k=0; k<100; k++);
	  // wait_us(1);		// dacout関数で全データ出力されるのを待つ
  }
	disable_timer( 0 );
//  LPC_GPIO1->FIOCLR |= (0x01 << 18);
  // 音が途切れることは無いのか？
  dac_e = dacbuf;  // potential thread problem ??  no...
  dac_s = dacbuf;

  while (nsamples--) {
	    signed int sample_l,sample_r;
	    sample_l = scale(*left_ch);
	    sample_r = scale(*right_ch);
		*dac_e = (unsigned short)(sample_l);
	    dac_e++;
		*dac_e = (unsigned short)(sample_r);
	    dac_e++;
	    left_ch++;
	    right_ch++;
  }
#endif

  return MAD_FLOW_CONTINUE;
}

/*
 * MONO化output関数
 */
static
enum mad_flow output_monauralize(void *data,
		     struct mad_header const *header,
		     struct mad_pcm *pcm)
{
	  unsigned int nsamples;
	  mad_fixed_t const *left_ch, *right_ch;
	  mad_fixed_t	mono_w;
	  static struct audio_dither dither;
	  dacout_s *dac_w;

	  /* pcm->samplerate contains the sampling frequency */
	  //nchannels = pcm->channels;
	  nsamples  = pcm->length;
	  left_ch   = pcm->samples[0];
	  right_ch  = pcm->samples[1];

	  // dac_s2が空くのを待つ(SysTick_Handler参照)
	  while(dac_s2 < dac_e2){
		  Pause();
	  }

	  dac_w = buftab[dacflg];
	  dac_s2 = dac_w;
	  while (nsamples--) {
		 mono_w = *left_ch / 2 + *right_ch / 2;
			*dac_w = (unsigned short)audio_linear_dither(16, mad_f_mul(mono_w, VolumeFactor), &dither);
			dac_w++;
	    	*dac_w = *(dac_w - 1);
			dac_w++;
			left_ch++;
			right_ch++;
	  }
	  dac_e2 = dac_w;
	  dacflg = 1 - dacflg;

	  return MAD_FLOW_CONTINUE;
}


/*
 * MONO用output関数
 */
static
enum mad_flow output_mono(void *data,
		     struct mad_header const *header,
		     struct mad_pcm *pcm)
{
  unsigned int nsamples;
  mad_fixed_t const *left_ch, *right_ch;
  static struct audio_dither dither;
  dacout_s *dac_w;

  //nchannels = pcm->channels;
  nsamples  = pcm->length;
  left_ch   = pcm->samples[0];
  right_ch  = pcm->samples[1];

  // dac_s2が空くのを待つ(SysTick_Handler参照)
  while(dac_s2 < dac_e2){
	  Pause();
  }

  dac_w = buftab[dacflg];
  dac_s2 = dac_w;
  while (nsamples--) {
		*dac_w = (unsigned short)audio_linear_dither(16, mad_f_mul(*left_ch, VolumeFactor), &dither);
	    dac_w++;
    	*dac_w = *(dac_w - 1);
	    dac_w++;
	    left_ch++;
	    right_ch++;
  }
  dac_e2 = dac_w;
  dacflg = 1 - dacflg;

  return MAD_FLOW_CONTINUE;
}


/*
 * This is the error callback function. It is called whenever a decoding
 * error occurs. The error is indicated by stream->error; the list of
 * possible MAD_ERROR_* errors can be found in the mad.h (or stream.h)
 * header file.
 */
static
enum mad_flow error(void *data,
		    struct mad_stream *stream,
		    struct mad_frame *frame)
{
#if !defined(MP3_12)
  struct buffer *buffer = data;

  fprintf(stderr, "decoding error 0x%04x (%s) at byte offset %u\n",
	  stream->error, mad_stream_errorstr(stream),
	  stream->this_frame - buffer->start);
#endif

  /* return MAD_FLOW_BREAK here to stop decoding (and propagate an error) */

  return MAD_FLOW_CONTINUE;
}


/*
 * This is the function called by main() above to perform all the decoding.
 * It instantiates a decoder object and configures it with the input,
 * output, and error callback functions above. A single call to
 * mad_decoder_run() continues until a callback function returns
 * MAD_FLOW_STOP (to stop decoding) or MAD_FLOW_BREAK (to stop decoding and
 * signal an error).
 * minimad.c(libmad)のdecode関数を元に作成。
 */
static
int decode()
{
//  struct buffer buffer;
  struct mad_decoder decoder;
  int result;

  /* initialize our private message structure */
//  buffer.start  = start;
//  buffer.length = length;

  /* configure input, output, and error functions */
  mad_decoder_init(&decoder, 0 /* &buffer */,
		   input, 0 /* header */, 0 /* filter */, output_monauralize,
		   error, 0 /* message */);

  /* start decoding */
  result = mad_decoder_run(&decoder, MAD_DECODER_MODE_SYNC);

  /* release the decoder */
  mad_decoder_finish(&decoder);

  return result;
}

/*
 * MP3ファイルごとの初期処理
 * 	output関数
 * 	I2S設定
 * 	SysTick設定
 */
int init_each_mp3(struct mad_decoder *decoder){
	struct mad_header *header;
	header = &(decoder->sync->frame.header);

	// decoder->output_funcの初期値はoutput_monauralize
    if (header->mode == MAD_MODE_SINGLE_CHANNEL){
    	decoder->output_func = output_mono;
    } else if (SpeakerOn == 0 || ChannelMode == 1){		// Speaker Off or Stereo
    	decoder->output_func = output;
    }
    if (samplerate != header->samplerate){	// 設定済みsamplerateと異なる時だけi2s_beginを実行する(POPノイズ対策)
    	if (SpeakerOn == 1){
    		LPC_GPIO0->FIOCLR |= (0x01 << SPEAKER_MUTE_BIT);	// Speaker off : P0.26 (set low)
    	}
    	i2s_begin(header->samplerate, 16);
    	samplerate = header->samplerate;
    	if (SpeakerOn == 1){
    		LPC_GPIO0->FIOSET |= (0x01 << SPEAKER_MUTE_BIT);	// speaker on : P0.26 (set high)
    	}
    } else {
    	i2s_start();
    }
	if (SysTick_Config(SystemCoreClock / header->samplerate - 1)) {
	}
	return MAD_FLOW_CONTINUE;
}

/*
 * SysTickタイマー　オフ
 */
void SysTick_Off(){
	SysTick->CTRL &= ~SysTick_CTRL_ENABLE_Msk;
}

/*
 * SysTickタイマー　オン
 */
void SysTick_On(){
	SysTick->CTRL |= SysTick_CTRL_ENABLE_Msk;
}

//  SysTick
void SysTick_Handler(void) {
	if(dac_s < dac_e){
		i2s_write(dac_s, 2);
		dac_s += 2;
	} else if (dac_s2 < dac_e2){
		dac_s = dac_s2;
		dac_e = dac_e2;
		i2s_write(dac_s, 2);
		dac_s += 2;
		dac_s2 = dac_e2;
	}
}

// TIMER0
void TIMER0_IRQHandler (void)
{
	LPC_TIM0->IR = 1;			/* clear interrupt flag */
	timer0_counter++;

	return;
}

void Pause(){
	timer0_counter = 0;
	while (timer0_counter == 0){

	}
}

// TIMER1
void TIMER1_IRQHandler (void){
	//static int	PinEventTmp;
	//static int	ShiftOnTmp;

	LPC_TIM1->IR = 1;			/* clear interrupt flag */
	timer1_counter++;
	if (timer1_counter >= TIMER1_1MIN){
		timer1_counter = 0;
		AutoPoffTimer++;
	}

#if 0
    if (CHATTERING_TMR == 0){

		if ((LPC_GPIO2->FIOPIN0 & (1 << 5)) == 0){			// P2.5
			if (NowPlaying == 0){
				PinEvent = SWUP_PRESSED;
				if ((LPC_GPIO0->FIOPIN0 & (1 << 4)) == 0) {			// P0.4 (Shift On)
					PinEvent |= SWSHIFT_MASK;
				}
			} else {	// 再生中は音量調整 (この関数内で処理してしまうのでPinEventを設定しない)
		    	if ((LPC_GPIO0->FIOPIN0 & (1 << 4)) != 0) {			// P0.4
					int		i;
					for (i=MIN_VOLUMETAB_IDX; i<=MAX_VOLUMETAB_IDX; i++){	// VolumeTabから直上の値を探す
						if (VolumeTab[i] > Volume){
							Volume = VolumeTab[i];
							break;
						}
					}
				} else {	// Shift+Upで音量微調整
					uint16_t	vstep;
					vstep = Volume / 10;
					if (vstep == 0) vstep = 1;
					if (Volume + vstep <= MAX_VOLUME){
						Volume += vstep;
						//VolumeFactor = mad_f_tofixed((long)Volume/(double)MAX_VOLUME);
					}
				}
				VolumeFactor = mad_f_tofixed((long)Volume/(double)MAX_VOLUME);
				CHATTERING_TMR = 1;
			}
		} else if ((LPC_GPIO2->FIOPIN0 & (1 << 4)) == 0){	// P2.4
			PinEvent = SWPREV_PRESSED;
		} else if ((LPC_GPIO2->FIOPIN0 & (1 << 3)) == 0){	// P2.3
			PinEvent = SWNEXT_PRESSED;
		} else if ((LPC_GPIO2->FIOPIN0 & (1 << 2)) == 0){	// P2.2
			if (NowPlaying == 0){
				PinEvent = SWDOWN_PRESSED;
				if ((LPC_GPIO0->FIOPIN0 & (1 << 4)) == 0) {			// P0.4
					PinEvent |= SWSHIFT_MASK;
				}
			} else {
		    	if ((LPC_GPIO0->FIOPIN0 & (1 << 4)) != 0) {			// P0.4
					int		i;
					for (i=MAX_VOLUMETAB_IDX; i>=MIN_VOLUMETAB_IDX; i--){
						if (VolumeTab[i] < Volume){
							Volume = VolumeTab[i];
							break;
						}
					}
				} else {
					uint16_t	vstep;
					vstep = Volume / 10;
					if (vstep == 0) vstep = 1;
					if (Volume >= vstep){
						Volume -= vstep;
						//VolumeFactor = mad_f_tofixed((long)Volume/(double)MAX_VOLUME);
					}
				}
				VolumeFactor = mad_f_tofixed((long)Volume/(double)MAX_VOLUME);
				CHATTERING_TMR = 1;
			}
		} else if ((LPC_GPIO0->FIOPIN1 & (1 << 3)) == 0) {	// P0.11
			PinEvent = SWDESELECT_PRESSED;
		} else if ((LPC_GPIO2->FIOPIN0 & (1 << 1)) == 0) {	// P2.1
			PinEvent = SWSELECT_PRESSED;
		} else if ((LPC_GPIO0->FIOPIN0 & (1 << 5)) == 0) {	// P0.5
			PinEvent = SWOPTION_PRESSED;
		} else if ((LPC_GPIO0->FIOPIN1 & (1 << 2)) == 0) {	// P0.10
			PinEvent = SWPLAY_PRESSED;
		} else if ((LPC_GPIO1->FIOPIN3 & (1 << 6)) == 0){	// P1.30
			PinEvent = SWPOFF_PRESSED;
		} else {
			PinEvent = 0;
		}

		if (PinEvent != 0){
			CHATTERING_TMR = 1;
		}
    } else if (CHATTERING_TMR > 0){
    	CHATTERING_TMR++;
    	if (CHATTERING_TMR == 3){
    		CHATTERING_TMR = 0;
    	}
    }
#endif

}

// 9/23
//// TIMER2
//void TIMER2_IRQHandler (void){
//	LPC_TIM2->IR = 1;			/* clear interrupt flag */
//	timer2_counter++;
//	onBoardLed1_On();
//	//GLCD_DISP_OF();
//	glcd_backlight_off();
//}

// ミリ秒単位の待ち時間
// TIMER0を使うので再生中は使えない。
void WaitMs(int ms){
	ms *= 10;
	timer0_counter = 0;
	while (timer0_counter < ms){

	}
}

#if 0
void test_read_write(){
//	FIL 		Fil1;
//	FIL 		Fil2;
	FIL 		*pFil1;
	FIL 		*pFil2;
    FRESULT     res;
    UINT		br;
    UINT		br2;
    char		buff[1024];

    pFil1 = (FIL *)mad_malloc(sizeof(FIL));
    pFil2 = (FIL *)mad_malloc(sizeof(FIL));

    res = f_open(pFil1 /*&Fil1*/, "/MSC1_in/test.dat", FA_READ | FA_OPEN_EXISTING);
    if (res != FR_OK){
//        DEBUG_MSG ("f_open failed.\n\r");
        return ;
    }
    res = f_open(pFil2 /*&Fil2*/, "/MSC1_out/testout.dat", FA_WRITE | FA_CREATE_ALWAYS);
    if (res != FR_OK){
//        DEBUG_MSG ("f_open failed.\n\r");
        return ;
    }
    do {
    	res = f_read(pFil1 /*&Fil1*/, buff, 1024, &br);
    	if (res != FR_OK || br == 0){
    		break;
    	}
    	//DEBUG_MSG(buff);
    	res = f_write(pFil2 /*&Fil2*/, buff, br, &br2);
		if (res != FR_OK){
			break;
		}
    } while (1);

    f_close(pFil1 /*&Fil1*/);
    f_close(pFil2 /*&Fil2*/);
//    DEBUG_MSG ("test_read_write OK.\n\r");
}

#endif
